import shutil
import os
import os.path as op
import tempfile
import logging as log

def get_validation_report(x, e, report_name):
    r = x.select.experiment(e).resource('BBRC_VALIDATOR')
    pdf = {each.label():each for each in list(r.files()) \
        if report_name in each.label() and \
        each.label().endswith('.pdf')}

    if not r.exists():
        log.error('%s has no BBRC_VALIDATOR resource'%e)
        return None
    if len(pdf.items()) == 0:
        log.error('%s has no %s'%(e, report_name))
        return None

    assert(len(list(pdf.keys())) == 1)
    f = pdf[list(pdf.keys())[0]]
    return f

def download_experiment(x, e, resource_name, validation_report, overwrite,
        destdir):
    if not resource_name is None:
        r = x.select.experiment(e).resource(resource_name)
        if not r.exists():
            log.error('%s has no %s resource'%(e, resource_name))
            return
        dd = op.join(destdir, e)
        if op.isdir(dd) and not overwrite:
            msg = '%s already exists. Skipping %s.'%(dd, e)
            log.error(msg)
        else:
            if op.isdir(dd) and overwrite:
                msg = '%s already exists. Overwriting %s.'%(dd, e)
                log.warning(msg)

            os.mkdir(dd)
            r.get(dest_dir=dd)

    f = get_validation_report(x, e, validation_report)
    if not f is None:
        fp = op.join(destdir, f.label()) if resource_name is None \
            else op.join(dd, f.label())
        if resource_name is None:
            log.debug('Saving it in %s.'%fp)
            f.get(dest=fp)



import logging as log
from datetime import datetime
import os.path as op
from bx.parse import Command

class NiftiCommand(Command):
    nargs = 1

    def __init__(self, *args, **kwargs):
        super(NiftiCommand, self).__init__(*args, **kwargs)
        s = input('Enter Sequence Name (i.e. Series Description):')
        self.sequence_name = s

    def parse(self, test=False):
        id = self.args[0]
        self.run_id(id, download_sequence, test=test,
            sequence_name=self.sequence_name, destdir=self.destdir)



def download_sequence(x, experiments, sequence_name, destdir):

    import os, shutil
    from glob import glob
    import tempfile
    from tqdm import tqdm
    columns = ['label', 'subject_label']

    for e in tqdm(experiments):
        log.debug('Experiment %s:'%e['ID'])


        scans = x.select.experiment(e['ID']).scans()
        seq_scans = [s for s in scans if not s.label().startswith('0-') and\
            sequence_name == s.attrs.get('type')]
        log.debug('Found following scans: %s'%seq_scans)

        for s in seq_scans:
            r = s.resource('NIFTI')
            if not r.exists():
                log.error('%s has no NIFTI'%e)
                return
            wd = tempfile.mkdtemp()

            r.get(dest_dir=wd, extract=True)
            fp = glob(op.join(wd, 'NIFTI', '*.nii.gz'))
            assert(len(fp)==1)
            fp = fp[0]
            fn = '%s_%s_%s_%s.nii.gz'%(e['subject_label'], e['label'], s.label(), e['ID'])
            log.debug('Saving %s...'%fn)
            shutil.move(fp, op.join(destdir, fn))
            shutil.rmtree(wd)



import logging as log
from datetime import datetime
import os.path as op
from bx.parse import Command

class IDCommand(Command):
    nargs = 1

    def __init__(self, *args, **kwargs):
        super(IDCommand, self).__init__(*args, **kwargs)

    def parse(self, test=False):
        id = self.args[0]
        df = self.run_id(id, get_id_table, test=test)

        self.to_excel(id, df)


def get_id_table(x, experiments):
    table = []
    from tqdm import tqdm
    columns = ['label', 'subject_label']
    for e in tqdm(experiments):
        exp = x.array.experiments(experiment_id=e['ID'], columns=columns).data[0]
        table.append([exp['ID'], exp['label'], exp['subject_label']])

    import pandas as pd
    columns = ['ID', 'label', 'subject_label']
    df = pd.DataFrame(table, columns=columns).set_index('ID').sort_index()
    return df



import pandas as pd
import json
import os.path as op
import os
import tempfile
import argparse
import pkgutil
import inspect
from datetime import datetime
from tqdm import tqdm

from .validation import validation_scores
from .download import *
from .dicom import *


def __get_modules__(m):
    modules = []
    prefix = m.__name__ + '.'
    log.info('prefix : %s'%prefix)
    for importer, modname, ispkg in pkgutil.iter_modules(m.__path__, prefix):
        module = __import__(modname , fromlist='dummy')
        if not ispkg:
            modules.append(module)
        else:
            modules.extend(__get_modules__(module))
    return modules

def __find_all_commands__(m):
    ''' Browses bx and looks for any class named as a Command'''
    modules = []
    classes = []
    modules = __get_modules__(m)
    forbidden_classes = [] #Test, ScanTest, ExperimentTest]
    for m in modules:
        for name, obj in inspect.getmembers(m):
            if inspect.isclass(obj) and 'Command' in name \
                    and not obj in forbidden_classes:
                classes.append(obj)
    return classes

class readable_dir(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        prospective_dir=values
        if not os.path.isdir(prospective_dir):
            msg = "readable_dir:{0} is not a valid path".format(prospective_dir)
            raise argparse.ArgumentTypeError(msg)
        if os.access(prospective_dir, os.R_OK):
            setattr(namespace,self.dest,prospective_dir)
        else:
            msg = "readable_dir:{0} is not a readable dir".format(prospective_dir)
            raise argparse.ArgumentTypeError(msg)

def check_xnat_item(a, x):
    projects = [e.label() for e in list(x.select.projects())]
    experiments = []
    for p in projects:
        exp = x.array.experiments(project_id=p).data
        experiments.extend([e['ID'] for e in exp])

    if a in projects:
        log.info('Project detected: %s'%a)
        return 0
    elif a in experiments:
        log.info('Experiment detected: %s'%a)
        return 1
    else:
        log.error('%s is not a project/experiment'%a)
        return -1


def download_experiments(x, id, resource_name, validation_report, destdir,
         overwrite=False, test=False):
    t = check_xnat_item(id, x)

    max_rows = 1 if test else None

    experiments = []
    if t == 1:
        experiments = [id]

    elif t == 0:
        for e in x.array.experiments(project_id=id,
                columns=['label']).data[:max_rows]:
            experiments.append(e['ID'])

    log.info('Now initiating download for %s experiments.'%len(experiments))
    for e in tqdm(experiments):
        log.debug(e)
        download_experiment(x, e, resource_name, validation_report,
            overwrite, destdir)


def download_measurements(x, func, id=None, test=False, **kwargs):

    t = check_xnat_item(id, x)
    max_rows = 25 if test else None

    experiments = []
    if t == 1:
        experiments = x.array.experiments(experiment_id=id,
            columns=['label', 'subject_label']).data

    elif t == 0:
        experiments = []
        for e in x.array.experiments(project_id=id,
                columns=['label', 'subject_label']).data[:max_rows]:
            experiments.append(e)
    else:
        print('%s is not a project or an experiment'%t)

    return func(x, experiments, **kwargs)



def parse(parser, command, test=False):
    c = command(parser.command, parser.args, parser.xnat, parser.destdir, parser.overwrite)

    if len(c.args) == 0:
        msg = 'display help message for %s'%command
        log.info(msg)

    elif len(c.args) == command.nargs:
        c.parse(test=test)

    else:
        msg = 'missing argument(s)'
        log.info(msg)



class Command(object):
    def __init__(self, command, args, xnat_instance, destdir, overwrite):
        self.command = command
        self.args = args
        self.xnat = xnat_instance
        self.destdir = destdir
        self.overwrite = overwrite

    def run_id(self, id, func, test=False, **kwargs):
        from bx import parse
        t = parse.check_xnat_item(id, self.xnat)

        max_rows = 1 if test else None
        columns = ['label', 'subject_label']
        experiments = []
        if t == 1:
            experiments = self.xnat.array.experiments(experiment_id=id,
                    columns=columns).data

        elif t == 0:
            for e in self.xnat.array.experiments(project_id=id,
                    columns=columns).data[:max_rows]:
                experiments.append(e)

        return func(self.xnat, experiments, **kwargs)

    def to_excel(self, id, df):
        dt = datetime.today().strftime('%Y%m%d_%H%M%S')
        fn = 'bx_%s_%s_%s.xlsx'%(self.command, id, dt)
        fp = op.join(self.destdir, fn)
        log.info('Saving it in %s'%fp)
        df.to_excel(fp)

def parse_args(command, args, x, destdir=tempfile.gettempdir(), overwrite=False,
        test=False):

    parser = Command(command, args, x, destdir, overwrite=overwrite)

    import bx
    commands = __find_all_commands__(bx)
    commands = {e.__name__.split('.')[-1].lower()[:-7]: e for e in commands}

    if command in commands.keys():
        log.debug('Command: %s'%command)
        if command == 'freesurfer6_hires':
            command = 'freesurfer6'
        parse(parser, commands[command], test=test)

    else:
        msg = '%s not found (valid commands: %s)'%(command, commands)
        log.info(msg)
        raise Exception(msg)


def create_parser():
    import argparse
    cfgfile = op.join(op.expanduser('~'), '.xnat.cfg')
    parser = argparse.ArgumentParser(description='bx')
    parser.add_argument('command', help='BX command')
    parser.add_argument('args', help='BX command', nargs="*")
    parser.add_argument('--config', help='XNAT configuration file',
        required=False, type=argparse.FileType('r'), default=cfgfile)
    parser.add_argument('--dest', help='Destination folder',
        required=False, action=readable_dir)
    parser.add_argument('--verbose', '-V', action='store_true', default=False,
        help='Display verbosal information (optional)', required=False)
    parser.add_argument('--overwrite', '-O', action='store_true', default=False,
        help='Overwrite', required=False)
    return parser



import logging as log
from datetime import datetime
import os.path as op
from bx.parse import Command


class MRDatesCommand(Command):
    nargs = 1

    def __init__(self, *args, **kwargs):
        super(MRDatesCommand, self).__init__(*args, **kwargs)

    def parse(self, test=False):
        id = self.args[0] #should be a project or an experiment_id
        from bx import parse
        t = parse.check_xnat_item(id, self.xnat)

        if t == 0:
            log.debug('Project detected: %s'%id)

            from bx.dicom import collect_mrdates
            df = collect_mrdates(self.xnat, id=id, test=test)
            if self.destdir == None:
                self.destdir = tempfile.gettempdir()
            self.to_excel(id, df)

        elif t == 1:
            log.debug('Experiment detected: %s'%id)
            from bx.dicom import collect_mrdates
            sd = collect_mrdates(self.xnat, id=id)
            print(sd)
            log.info('Scan date: %s'%sd)

        else:
            log.error('No project/experiment found: %s'%id)



import logging as log

def get_scandate(experiment_id, x, t1_scan_label='T1_ALFA1'):

    columns = ['xsiType', 'xnat:imagescandata/type', 'xnat:imagescandata/ID']
    scans = x.array.scans(experiment_id=experiment_id, columns=columns).data
    t1_scans = {e['xnat:imagescandata/id']:e for e in scans \
        if e['xnat:imagescandata/type'] == t1_scan_label}

    if len(t1_scans.items()) == 0:
        msg = 'No T1 found for %s: %s. Trying with all of them.'\
            %(experiment_id, [e['xnat:imagescandata/id'] for e in scans])
        log.warning(msg)
        t1_scans = {e['xnat:imagescandata/id']:e for e in scans \
            if not e['xnat:imagescandata/id'].startswith('OT-')\
            and not e['xnat:imagescandata/id'].startswith('O-')}


    max_nb = sorted(t1_scans.keys())[-1]
    log.debug('Found scan: %s'%max_nb)
    scan = x.select.experiment(experiment_id).scan(max_nb)

    f = list(scan.resource('DICOM').files())[0]

    import pydicom
    import tempfile
    import os.path as op
    fp = op.join(tempfile.gettempdir(), 'test.dcm')
    f.get(dest=fp)
    d = pydicom.read_file(fp)

    if hasattr(d, 'AcquisitionDate'):
        acquisition_date = d.AcquisitionDate
    else:
        acquisition_date = d.AcquisitionDateTime[:8]

    import os
    os.remove(fp)
    return acquisition_date


def collect_mrdates(x, id, overwrite=False, test=False):

    def __create_table__(data):
        import pandas as pd
        df = pd.DataFrame(data, columns=('ID', 'label', 'subject_label', 'scandate'))
        df['scandate'] = pd.to_datetime(df['scandate'])
        df = df.set_index('ID').sort_index()
        return df

    from tqdm import tqdm
    from bx import parse
    t = parse.check_xnat_item(id, x)

    experiments = []
    columns = ['label', 'subject_ID', 'subject_label']
    if t == 1:
        experiments = x.array.experiments(experiment_id=id, columns=columns).data

    elif t == 0:
        max_rows = 25 if test else None
        for e in x.array.experiments(project_id=id, columns=columns).data[:max_rows]:
            experiments.append(e)
    else:
        print('%s is not a project or an experiment'%t)

    data = []
    for e in tqdm(experiments):
        try:
            log.debug('Experiment ID: %s Subject label: %s'%(e['ID'], e['subject_label']))
            row = [e['ID'], e['label'], e['subject_label']]
            d = get_scandate(e['ID'], x)
            row.append(d)
            data.append(row)
        except KeyboardInterrupt:
            return __create_table__(data)
        except Exception as exc:
            log.error('Failed with %s. Skipping it. (%s)'%(e['ID'], exc))
            continue
    return __create_table__(data)



from datetime import datetime
import os.path as op
import logging as log
from bx.parse import Command


class FreeSurfer6Command(Command):
    nargs = 2
    def __init__(self, *args, **kwargs):
        super(FreeSurfer6Command, self).__init__(*args, **kwargs)

    def parse(self, test=False):
        subcommand = self.args[0]
        id = self.args[1] #should be a project or an experiment_id
        print(id)
        if subcommand in ['aparc', 'aseg', 'hippoSfVolumes']:
            self.subcommand_aparc(test)

        elif subcommand in ['files', 'report']:
            self.subcommand_files_or_report(test)

        elif subcommand == 'tests':
            from bx.validation import subcommand_tests
            subcommand_tests(parser=self, test=test, validation='FreeSurferValidator',
                id=id, version=['##0390c55f', '4e37c9d0'])

    def subcommand_aparc(self, test=False):
        subcommand = self.args[0]
        id = self.args[1]

        known_suffixes = ['_HIRES', '_SUBFIELDS']
        resource_name = 'FREESURFER6'
        suffix = ''
        if '_' in self.command:
            suffix = '_%s'%self.command.split('_')[1].upper()
            if not suffix in known_suffixes:
                print('%s not known (known suffixes: %s)'
                    %(self.command, known_suffixes))
            else:
                resource_name = 'FREESURFER6%s'%suffix

        from bx import parse
        if subcommand == 'aparc':
            df = parse.download_measurements(self.xnat, aparc_measurements,  id, test, resource_name=resource_name)
        elif subcommand == 'aseg':
            df = parse.download_measurements(self.xnat, aseg_measurements, id, test, resource_name=resource_name)
        elif subcommand == 'hippoSfVolumes':
            df = parse.download_measurements(self.xnat, hippoSfVolumes_measurements, id, test, resource_name=resource_name)

        dt = datetime.today().strftime('%Y%m%d_%H%M%S')
        fn = 'bx_%s_%s_%s_%s.xlsx'%(self.command, subcommand, id, dt)
        fp = op.join(self.destdir, fn)
        log.info('Saving it in %s'%fp)
        df.to_excel(fp)


    def subcommand_files_or_report(self, test=False):
        subcommand = self.args[0]
        id = self.args[1]

        known_suffixes = ['_HIRES', '_SUBFIELDS']
        resource_name = 'FREESURFER6'
        suffix = ''
        if '_' in self.command:
            suffix = '_%s'%self.command.split('_')[1].upper()
            if not suffix in known_suffixes:
                print('%s not known (known suffixes: %s)'
                    %(self.command, known_suffixes))
            else:
                resource_name = 'FREESURFER6%s'%suffix


        report_only = subcommand == 'report'
        validation_report = 'SPM12SegmentValidator'
        suffix = 'Hires' if resource_name.endswith('_HIRES') else ''
        validation_report = 'FreeSurfer%sValidator'%suffix
        if report_only:
            resource_name = None

        from bx import parse
        parse.download_experiments(self.xnat, id, resource_name,
            validation_report, self.destdir, self.overwrite,
            test=test)

def freesurfer6_measurements(x, func, experiments, resource_name='FREESURFER6'):
    from tqdm import tqdm
    import pandas as pd

    table = []
    for e in tqdm(experiments):
        log.debug(e)
        try:
            s = e['subject_label']
            r = x.select.experiment(e['ID']).resource(resource_name)
            if not r.exists():
                log.error('%s has no %s resource'%(e, resource_name))
                continue
            if func == 'aparc':
                volumes = r.aparc()
            elif func == 'aseg':
                volumes = r.aseg()
            elif func == 'hippoSfVolumes':
                volumes = r.hippoSfVolumes(mode='T1')
            volumes['subject'] = s
            volumes['ID'] = e['ID']
            table.append(volumes)
        except KeyboardInterrupt:
            return pd.concat(table).set_index('ID').sort_index()
        except Exception as exc:
            log.error('Failed for %s. Skipping it.'%e)
            log.error(exc)
            continue
    hippoSfVolumes = pd.concat(table).set_index('ID').sort_index()
    return hippoSfVolumes

def aparc_measurements(x, experiments, resource_name='FREESURFER6'):
    return freesurfer6_measurements(x, 'aparc', experiments, resource_name=resource_name)

def aseg_measurements(x, experiments, resource_name='FREESURFER6'):
    return freesurfer6_measurements(x, 'aseg', experiments, resource_name=resource_name)

def hippoSfVolumes_measurements(x, experiments, resource_name='FREESURFER6'):
    return freesurfer6_measurements(x, 'hippoSfVolumes', experiments, resource_name=resource_name)



import logging as log
import os.path as op
from datetime import datetime
from tqdm import tqdm

def collect_reports(xnat_instance, experiments,
        validator_name='ArchivingValidator', version=['toto']):
    import json
    url = '/data/experiments/%s/resources/BBRC_VALIDATOR/files/%s'
    reports = {}

    for e in tqdm(experiments):
        columns = ['ID', 'label', 'xsiType']
        exp = xnat_instance.array.experiments(experiment_id=e,
                                              columns=columns).data
        assert(len(exp)==1)
        uri = url%(exp[0]['ID'], '%s_%s.json'%(validator_name, exp[0]['label']))
        r = xnat_instance.select.experiment(e).resource('BBRC_VALIDATOR')
        if not r.exists(): continue
        f = r.file('%s_%s.json'%(validator_name, exp[0]['label']))
        if not f.exists(): continue

        j = json.loads(xnat_instance.get(f._uri).text)
        if 'version' not in j.keys():
            log.warning('Version not found in report %s'%j.keys())
            continue
        if j['version'] not in version: continue
        fields = list(j.keys())
        try:
            for each in ['version', 'generated', 'experiment_id']:
                fields.remove(each)
        except ValueError:
            msg = 'No valid report found (%s).'%e
            log.error(msg)
            raise Exception(msg)
        if j['version'] not in  version: continue
        reports[e] = j

    return reports

def validation_scores(x, validator, version,  id, test=False):
    from bx import parse
    t = parse.check_xnat_item(id, x)

    experiments = []
    if t == 1:
        experiments = [id]

    elif t == 0:
        max_rows = 1 if test else None
        experiments = []
        for e in x.array.experiments(project_id=id,
                columns=['label']).data[:max_rows]:
            experiments.append(e['ID'])

    res = []
    fields = []
    log.info('Looking for experiments with %s report with versions %s.'\
                %(validator, version))
    reports = dict(list(collect_reports(x, validator_name=validator,
        experiments=experiments, version=version).items())[:max_rows])
    print(reports)
    log.info('Now initiating download for %s experiment(s).'\
            %len(reports.items()))

    for e, report in tqdm(reports.items()):
        fields = list(report.keys())
        for each in ['version', 'generated', 'experiment_id']:
            fields.remove(each)
        row = [e]
        row.extend([report[f]['has_passed'] for f in fields])
        res.append(row)

    import pandas as pd
    fields.insert(0, 'ID')
    df = pd.DataFrame(res, columns=fields).set_index('ID')
    return df


def subcommand_tests(parser, validator, version, test=False):
    subcommand = parser.args[0]
    id = parser.args[1]

    from bx.validation import validation_scores
    df = validation_scores(parser.xnat, validator=validator,
        id=id, version=version, test=test)

    dt = datetime.today().strftime('%Y%m%d_%H%M%S')
    fn = 'bx_%s_%s_%s_%s.xlsx'%(parser.command, subcommand, id, dt)
    fp = op.join(parser.destdir, fn)
    log.info('Saving it in %s'%fp)
    df.to_excel(fp)



import logging as log
from datetime import datetime
import os.path as op
from bx.parse import Command


class SPM12Command(Command):
    nargs = 2

    def __init__(self, *args, **kwargs):
        super(SPM12Command, self).__init__(*args, **kwargs)

    def parse(self, test=False):
        subcommand = self.args[0]
        id = self.args[1]

        if subcommand in ['files', 'report']:
            self.subcommand_files_or_report(test)

        elif subcommand == 'volumes':
            self.subcommand_volumes(test)

        elif subcommand == 'tests':
            from bx.validation import subcommand_tests
            subcommand_tests(parser=self, test=test, validator='SPM12SegmentValidator',
                version=['##0390c55f', '2bc4d861'])



    def subcommand_files_or_report(self, test=False):
        subcommand = self.args[0]
        id = self.args[1]

        report_only = subcommand == 'report'
        resource_name = 'SPM12_SEGMENT'
        validation_report = 'SPM12SegmentValidator'
        if report_only:
            resource_name = None

        from bx import parse
        parse.download_experiments(self.xnat, id, resource_name,
            validation_report, self.destdir, self.overwrite, test=test)


    def subcommand_volumes(self, test=False):

        subcommand = self.args[0]
        id = self.args[1]

        from bx import parse
        df = parse.download_measurements(self.xnat, spm12_volumes, id, test, resource_name='SPM12_SEGMENT')

        dt = datetime.today().strftime('%Y%m%d_%H%M%S')
        fn = 'bx_%s_%s_%s_%s.xlsx'%(self.command, subcommand, id, dt)
        fp = op.join(self.destdir, fn)
        log.info('Saving it in %s'%fp)
        df.to_excel(fp)


def spm12_volumes(x, experiments, resource_name):
    from tqdm import tqdm
    import pandas as pd

    table = []

    for e in tqdm(experiments):
        log.debug(e)
        try:
            r = self.xnat.select.experiment(e).resource(resource_name)
            if not r.exists():
                log.error('%s has no %s resource'%(e, resource_name))
                continue
            vols = [e]
            for kls in ['c1', 'c2', 'c3']:
                f = [each for each in r.files() if each.id().startswith(kls)][0]
                fp = tempfile.mkstemp('.nii.gz')[1]
                f.get(fp)
                d = nib.load(fp)
                size = np.prod(d.header['pixdim'].tolist()[:4])
                v = np.sum(d.dataobj) * size
                os.remove(fp)
                vols.append(v)
            table.append(vols)

        except KeyboardInterrupt:
            return pd.DataFrame(table, columns=['ID', 'c1', 'c2', 'c3']).set_index('ID').sort_index()
        except Exception as exc:
            log.error('Failed for %s. Skipping it.'%e)
            log.error(exc)
            continue

    df = pd.DataFrame(table, columns=['ID', 'c1', 'c2', 'c3']).set_index('ID').sort_index()
    return df



from bx.parse import Command

class ANTSCommand(Command):
    nargs = 2

    def __init__(self, *args, **kwargs):
        super(ANTSCommand, self).__init__(*args, **kwargs)

    def parse(self, test=False):
        
        subcommand = self.args[0]
        id = self.args[1] #should be a project or an experiment_id

        report_only = subcommand == 'report'
        resource_name = 'ANTS'
        validation_report = 'ANTSValidator'
        if report_only:
            resource_name = None

        from bx import download
        download.download_experiments(self.xnat, id, resource_name,
            validation_report, self.dest_dir, self.overwrite,
            test=test)
